# -*- coding: utf-8 -*-
from django.contrib import admin, messages
from django.utils.translation import ugettext_lazy as _
from django.contrib.admin.util import unquote
from django.http import HttpResponseRedirect
from .abstract_models import NotEnoughMoneyError


class BaseAffiliateAdmin(admin.ModelAdmin):
    list_display = ("aid", "balance", "total_payed", "total_payments_count",
        "created_at")


class BaseAffiliateStatsAdmin(admin.ModelAdmin):
    list_display = ("affiliate", "date", "total_views", "unique_visitors",
        "payments_count", "payments_amount", "rewards_amount")


class BaseAffiliateBannerAdmin(admin.ModelAdmin):
    list_display = "image", "get_width", "get_height", "enabled"
    list_editable = "enabled",

    def get_width(self, obj):
        return obj.image.width
    get_width.short_description = _("Width")

    def get_height(self, obj):
        return obj.image.height
    get_height.short_description = _("Height")


class BaseWithdrawRequestAdmin(admin.ModelAdmin):
    list_display = "affiliate", "amount", "status", "created_at"
    change_form_template = 'admin/request_change_form.html'
    readonly_fields = "payed_at", "status", "amount",

    def change_view(self, request, object_id, form_url='', extra_context=None):
        extra_context = extra_context or {}
        resp = super(BaseWithdrawRequestAdmin, self).change_view(
            request, object_id, form_url=form_url, extra_context=extra_context)
        if request.method == 'POST' and '_affiliate_payed' in request.POST\
                and isinstance(resp, HttpResponseRedirect):
            # assert, that HttpResponseRedirect means successfull operation
            payment_request = self.get_object(request, unquote(object_id))
            try:
                payment_request.payment_made()
                msg = _('Payment successfully marked as payed')
                msg_level = messages.INFO
            except NotEnoughMoneyError:
                msg = _('Payment failed: not enough money in affiliate account')
                msg_level = messages.ERROR
            self.message_user(request, msg, level=msg_level)
        return resp



# -*- coding: utf-8 -*-
from django.db import models
from .queryset import AffiliateStatsQuerySet, AffiliateBannerQuerySet,\
    WithdrawRequestQuerySet


class AffiliateStatsManager(models.Manager):
    def get_query_set(self):
        return AffiliateStatsQuerySet(self.model)

    def __getattr__(self, attr, *args):
        # see https://code.djangoproject.com/ticket/15062 for details
        if attr.startswith("_"):
            raise AttributeError
        return getattr(self.get_query_set(), attr, *args)


class AffiliateBannerManager(models.Manager):
    def get_query_set(self):
        return AffiliateBannerQuerySet(self.model)

    def __getattr__(self, attr, *args):
        # see https://code.djangoproject.com/ticket/15062 for details
        if attr.startswith("_"):
            raise AttributeError
        return getattr(self.get_query_set(), attr, *args)


class WithdrawRequestManager(models.Manager):
    def get_query_set(self):
        return WithdrawRequestQuerySet(self.model)

    def __getattr__(self, attr, *args):
        # see https://code.djangoproject.com/ticket/15062 for details
        if attr.startswith("_"):
            raise AttributeError
        return getattr(self.get_query_set(), attr, *args)



from __future__ import unicode_literals
from django.template.base import Library, TemplateSyntaxError, kwarg_re
from django.template.defaulttags import URLNode
from affiliate.tools import add_affiliate_code

register = Library()


class AffiliateURLNode(URLNode):
    def render(self, context, *args, **kwargs):
        url = super(AffiliateURLNode, self).render(context, *args, **kwargs)
        aid_code = context.get('affiliate_code', None)
        if aid_code:
            url = add_affiliate_code(url, aid_code)
        return url


@register.tag
def url_aff(parser, token):
    bits = token.split_contents()
    if len(bits) < 2:
        raise TemplateSyntaxError("'%s' takes at least one argument"
                                  " (path to a view)" % bits[0])
    try:
        viewname = parser.compile_filter(bits[1])
    except TemplateSyntaxError as exc:
        exc.args = (exc.args[0] + ". "
                "The syntax of 'url' changed in Django 1.5, see the docs."),
        raise
    args = []
    kwargs = {}
    asvar = None
    bits = bits[2:]
    if len(bits) >= 2 and bits[-2] == 'as':
        asvar = bits[-1]
        bits = bits[:-2]

    if len(bits):
        for bit in bits:
            match = kwarg_re.match(bit)
            if not match:
                raise TemplateSyntaxError("Malformed arguments to url tag")
            name, value = match.groups()
            if name:
                kwargs[name] = parser.compile_filter(value)
            else:
                args.append(parser.compile_filter(value))

    return AffiliateURLNode(viewname, args, kwargs, asvar)


@register.simple_tag(takes_context=True)
def render_banner(context, affiliate, banner, autoescape=False):
    img_html = affiliate.render_img(banner, context['request'],
        autoescape=autoescape)
    return img_html



# -*- coding: utf-8 -*-
from affiliate.tools import get_affiliate_param_name


def common(request):
    aff_param_name = get_affiliate_param_name()
    affiliate_code = request.GET.get(aff_param_name, None)
    return {'affiliate_code': affiliate_code}



from datetime import datetime
import logging

from django.conf import settings
from django.http import HttpResponseRedirect
from django.core.cache import get_cache
from .tools import get_affiliate_param_name, remove_affiliate_code,\
    get_seconds_day_left, get_affiliate_model, get_affiliatestats_model
from relish.helpers.request import get_client_ip

l = logging.getLogger(__name__)


AFFILIATE_NAME = get_affiliate_param_name()
AFFILIATE_SESSION = getattr(settings, 'AFFILIATE_SESSION', True)
AFFILIATE_SESSION_AGE = getattr(settings, 'AFFILIATE_SESSION_AGE', 5 * 24 * 60 * 60)
AFFILIATE_SKIP_PATH = getattr(settings, 'AFFILIATE_SKIP_PATH_STARTS', [])

C_PFX = 'a_'

AffiliateModel = get_affiliate_model()
AffiliateModelStats = get_affiliatestats_model()


class AffiliateMiddleware(object):
    datetime_format = '%Y-%m-%d %H:%M:%S'

    def process_request(self, request):
        aid = None
        session = request.session
        now = datetime.now()
        if request.method == 'GET':
            aid = request.GET.get(AFFILIATE_NAME, None)
            if aid:
                request.aid = aid
                if AFFILIATE_SESSION:
                    session['aid'] = aid
                    session['aid_dt'] = now.strftime(self.datetime_format)
                    url = remove_affiliate_code(request.get_full_path())
                    return HttpResponseRedirect(url)
        if not aid and AFFILIATE_SESSION:
            aid = session.get('aid', None)
            if aid:
                aid_dt = session.get('aid_dt', None)
                if aid_dt is None:
                    l.error('aid_dt not found in session')
                else:
                    aid_dt = datetime.strptime(aid_dt, self.datetime_format)
                    if (now - aid_dt).seconds > AFFILIATE_SESSION_AGE:
                        # aid expired
                        aid = None
                        session.pop('aid')
                        session.pop('aid_dt')
        request.aid = aid

    def process_response(self, request, response):
        aid = getattr(request, "aid", None)
        if not aid:
            l.error("aid not set")
        elif response.status_code == 200 and self.is_track_path(request.path):
            now = datetime.now()
            ip = get_client_ip(request)
            cache = get_cache('default')
            c_key = "".join((C_PFX, aid))
            ip_new, aid_ip_pool = self.is_new_ip(c_key, cache, ip)
            if ip_new:
                aid_ip_pool.add(ip)
                timeout = get_seconds_day_left(now)
                cache.set(c_key, aid_ip_pool, timeout)
            nb = AffiliateModelStats.objects.incr_count_views(aid, now,
                ip_new=ip_new)
            if not nb:
                try:
                    aff = AffiliateModel.objects.get(aid=aid)
                    AffiliateModelStats.objects.create(affiliate=aff,
                        total_views=1, unique_visitors=1)
                except AffiliateModel.DoesNotExist:
                    l.warning("Access with unknown affiliate code: {0}"
                        .format(aid))
        return response

    def is_track_path(self, path):
        return len(filter(path.startswith, AFFILIATE_SKIP_PATH)) == 0

    def is_new_ip(self, c_key, cache, ip):
        aid_ip_pool = cache.get(c_key)
        ip_new = True
        if aid_ip_pool:
            if isinstance(aid_ip_pool, set):
                ip_new = ip not in aid_ip_pool
        if not aid_ip_pool:
            aid_ip_pool = set()
        return ip_new, aid_ip_pool

# TODO: attach lazy method to request: affiliate, that return Affiliate instance



# -*- coding: utf-8 -*-
from decimal import Decimal as D
from django.views.generic import FormView
from django.utils.translation import ugettext_lazy as _
from django.conf import settings
from relish.views.messages import SuccessMessageMixin
from relish.decorators import instance_cache
from .tools import get_affiliate_model
from .forms import AffiliateCreateForm, AffiliateWithdrawRequestForm

MIN_REQUEST_AMOUNT = getattr(settings, 'AFFILIATE_MIN_BALANCE_FOR_REQUEST', D('1.0'))


class AffiliateBaseView(SuccessMessageMixin, FormView):
    """
    Probably you need to redefine some methods here to respect
    your custom Affiliate model.
    This view asserts, that Affiliate has OneToOne relationship to User.
    """
    template_name = "partner/affiliate.html"

    @property
    @instance_cache
    def user(self):
        return self.request.user

    def get_affiliate_model(self):
        return get_affiliate_model()

    def get_affiliate_banner_model(self):
        raise NotImplementedError()

    @property
    @instance_cache
    def affiliate(self):
        aff_model = self.get_affiliate_model()
        try:
            return aff_model.objects.get(user=self.user)
        except aff_model.DoesNotExist:
            return None

    def get_form_class(self):
        if self.affiliate:
            form_class = AffiliateWithdrawRequestForm
        else:
            form_class = AffiliateCreateForm
        return form_class

    def get_form_kwargs(self):
        kwargs = super(AffiliateBaseView, self).get_form_kwargs()
        affiliate = self.affiliate
        if affiliate:
            kwargs['affiliate'] = affiliate
        else:
            kwargs['user'] = self.user
        return kwargs

    def get_success_url(self):
        return self.request.get_full_path()

    def get_success_message(self):
        if self.affiliate:
            return _("Request for payment was sent")
        else:
            return _("Affiliate account successfully created")

    def form_valid(self, form):
        form.save()
        return super(AffiliateBaseView, self).form_valid(form)

    def get_context_data(self, **kwargs):
        context = super(AffiliateBaseView, self).get_context_data(**kwargs)
        affiliate = self.affiliate
        context['affiliate'] = affiliate
        if affiliate:
            context['min_request_amount'] = MIN_REQUEST_AMOUNT
            context['currency_label'] = affiliate.get_currency()
            context['requested'] = affiliate.pay_requests.pending()
            context['avaliable_for_request'] = affiliate.balance >= MIN_REQUEST_AMOUNT
            context['pay_requests'] = affiliate.pay_requests.all()
            aff_banner_model = self.get_affiliate_banner_model()
            context['banners'] = aff_banner_model.objects.enabled()
            context['visitor_stats'] = affiliate.stats.for_last_days(30)
        return context



# -*- coding: utf-8 -*-
from urlparse import urlparse, parse_qs
from urllib import urlencode
from datetime import datetime
from django.conf import settings
from django.db.models.loading import get_model


AFFILIATE_MODEL = settings.AFFILIATE_MODEL
AFFILIATE_COUNT_MODEL = settings.AFFILIATE_COUNT_MODEL


def get_affiliate_model():
    return get_model(*AFFILIATE_MODEL.split("."))


def get_affiliatestats_model():
    return get_model(*AFFILIATE_COUNT_MODEL.split("."))


def get_affiliate_param_name():
    return getattr(settings, "AFFILIATE_PARAM_NAME", 'aid')


def add_affiliate_code(url, aid_code):
    parsed = urlparse(str(url))
    params = parse_qs(parsed.query)
    aff_param_name = get_affiliate_param_name()
    params.update({aff_param_name: [str(aid_code)]})
    return "?".join((parsed.path, urlencode(params, doseq=True)))


def remove_affiliate_code(url):
    parsed = urlparse(str(url))
    params = parse_qs(parsed.query)
    aff_param_name = get_affiliate_param_name()
    params.pop(aff_param_name)
    return "?".join((parsed.path, urlencode(params, doseq=True)))


def get_seconds_day_left(now=None):
    now = now or datetime.now()
    end_day = datetime(year=now.year, month=now.month, day=now.day, hour=23,
        minute=59, second=59)
    return (end_day-now).seconds



# -*- coding: utf-8 -*-
from django import forms
from django.utils.translation import ugettext_lazy as _
from .tools import get_affiliate_model
AffiliateModel = get_affiliate_model()


class AffiliateCreateForm(forms.Form):
    """
    Redefine it to respect your custom Affiliate model.
    This form asserts, that Affiliate has OneToOne relationship to User.
    """
    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop('user', None)
        super(AffiliateCreateForm, self).__init__(*args, **kwargs)

    def clean(self):
        try:
            self.user.affiliate
            raise forms.ValidationError(_("Affiliate already created"))
        except AffiliateModel.DoesNotExist:
            pass
        return self.cleaned_data

    def save(self):
        aff = AffiliateModel.create_affiliate(user=self.user)
        return aff


class AffiliateWithdrawRequestForm(forms.Form):
    """
    Redefine it to respect your custom WithdrawRequest model.
    """
    def __init__(self, *args, **kwargs):
        self.affiliate = kwargs.pop('affiliate', None)
        super(AffiliateWithdrawRequestForm, self).__init__(*args, **kwargs)

    def clean(self):
        if self.affiliate is None:
            raise forms.ValidationError(_("Affiliate not found"))
        if self.affiliate.pay_requests.pending().exists():
            raise forms.ValidationError(_("Request is already sent"))
        return self.cleaned_data

    def save(self):
        return self.affiliate.create_payment_request()



import os
from setuptools import setup, find_packages
from affiliate import __author__, __version__


def __read(fname):
    try:
        return open(os.path.join(os.path.dirname(__file__), fname)).read()
    except IOError:
        return ''


install_requires = __read('requirements.txt').split()

setup(
    name='django-affiliate',
    author=__author__,
    author_email='alexevseev@gmail.com',
    version=__version__,
    description='Affiliate system for django',
    long_description=__read('README.rst'),
    platforms=('Any'),
    packages=find_packages(),
    install_requires=install_requires,
    keywords='django affiliate urls track'.split(),
    include_package_data=True,
    license='BSD License',
    package_dir={'affiliate': 'affiliate'},
    url='https://github.com/st4lk/django-affiliate',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Utilities',
    ],
)



# -*- coding: utf-8 -*-
from datetime import timedelta
from django.db.models.query import QuerySet
from django.db.models import F
from django.utils.timezone import now


class AffiliateStatsQuerySet(QuerySet):

    def incr_count_views(self, aid, date, ip_new=False):
        kw = dict(total_views=F('total_views')+1)
        if ip_new:
            kw['unique_visitors'] = F('unique_visitors')+1
        return self.filter(affiliate=aid, date=date).update(**kw)

    def for_last_days(self, days):
        # TODO: for MySQL add FORCE INDEX (affiliate_id)
        days_ago = now() - timedelta(days=days)
        return self.filter(date__gte=days_ago).order_by("-date")


class WithdrawRequestQuerySet(QuerySet):
    def pending(self):
        return self.filter(status=self.model.PAY_STATUS.pending)


class AffiliateBannerQuerySet(QuerySet):

    def enabled(self):
        return self.filter(enabled=True)



import django.dispatch


affiliate_post_reward = django.dispatch.Signal(
    providing_args=["affiliate", "amount"])
affiliate_post_withdraw = django.dispatch.Signal(
    providing_args=['payment_request'])



# -*- coding: utf-8 -*-
from decimal import Decimal as D
from django.utils.timezone import now
from django.db import models
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django.contrib.sites.models import get_current_site
from django.utils.html import escape
from model_utils import Choices
from relish.decorators import instance_cache
from .managers import AffiliateStatsManager, AffiliateBannerManager,\
    WithdrawRequestManager
from .tools import get_affiliate_param_name
from .signals import affiliate_post_withdraw, affiliate_post_reward

START_AID = getattr(settings, 'AFFILIATE_START_AID', "1000")
AID_NAME = get_affiliate_param_name()
BANNER_FOLDER = getattr(settings, 'AFFILIATE_BANNER_PATH', 'affiliate')
REWARD_AMOUNT = getattr(settings, 'AFFILIATE_REWARD_AMOUNT', D('5.0'))
REWARD_PERCENTAGE = getattr(settings, 'AFFILIATE_REWARD_PERCENTAGE', True)
TWOPLACES = D('0.01')


class NotEnoughMoneyError(Exception):
    pass


class AbstractAffiliateStats(models.Model):
    # don't create additional index on fk, as we've already declared
    # compound index, that is started with affiliate
    affiliate = models.ForeignKey(settings.AFFILIATE_MODEL, db_index=False,
        verbose_name=_("Affiliate"), related_name='stats')
    unique_visitors = models.IntegerField(_("Unique visitors count"),
        default=0)
    total_views = models.IntegerField(_("Total page views count"),
        default=0)
    payments_count = models.IntegerField(_("Number of payments"), default=0)
    payments_amount = models.DecimalField(_("Payments amount"),
        max_digits=6, decimal_places=2, default=D("0.0"))
    rewards_amount = models.DecimalField(_("Rewards amount"), max_digits=6,
        decimal_places=2, default=D("0.0"))
    date = models.DateField(_("Date"), auto_now_add=True)

    objects = AffiliateStatsManager()

    def __unicode__(self):
        return u"{0}, {1}".format(self.affiliate_id, self.date)

    class Meta:
        abstract = True
        # TODO would be great to have following fields as compound primary key
        # look https://code.djangoproject.com/wiki/MultipleColumnPrimaryKeys
        # and https://code.djangoproject.com/ticket/373
        unique_together = (
            ("affiliate", "date"),
        )
        verbose_name = _("Affiliate statistic")
        verbose_name_plural = _("Affiliate statistics")
        ordering = "-id",

    def incr_payments(self, purchase_total_price, reward, commit=True):
        self.payments_count += 1
        self.payments_amount += purchase_total_price
        self.rewards_amount += reward
        if commit:
            self.save()


class AbstractAffiliate(models.Model):
    aid = models.CharField(_("Affiliate code"), max_length=150,
        unique=True, primary_key=True)
    total_payments_count = models.IntegerField(
        _("Total attracted payments count"), default=0)
    total_payments_amount = models.DecimalField(
        _("Total attracted payments amount"), max_digits=6, decimal_places=2,
        default=D("0.0"))
    total_payed = models.DecimalField(_("Total payed to affiliate"),
        max_digits=6, decimal_places=2, default=D("0.0"))
    balance = models.DecimalField(_("Current balance"), max_digits=6,
        decimal_places=2, default=D("0.0"))
    created_at = models.DateTimeField(_("Created at"), auto_now_add=True)
    reward_amount = models.DecimalField(_("Reward amount"), max_digits=5,
        decimal_places=2, default=REWARD_AMOUNT)
    reward_percentage = models.BooleanField(_('In percent'),
        default=REWARD_PERCENTAGE)

    class Meta:
        abstract = True
        verbose_name = _("Affiliate")
        verbose_name_plural = _("Affiliates")

    def __unicode__(self):
        return self.aid

    def generate_aid(self):
        try:
            last_aid = self._default_manager.order_by("-aid")[0].aid
        except IndexError:
            last_aid = START_AID
        return str(int(last_aid) + 1)

    @instance_cache
    def render_link(self, request=None):
        site = self.get_site(request)
        return 'http://{domain}?{aid_name}={aid_code}'\
            .format(domain=site.domain, aid_name=AID_NAME, aid_code=self.aid)

    def render_html_a(self, request=None):
        site = self.get_site(request)
        link = self.render_link(request)
        return '<a href="{link}">{name}</a>'.format(link=link, name=site.name)

    def render_img(self, banner, request=None, autoescape=False):
        domain = self.get_site(request).domain.rstrip('/')
        link = self.render_link(request)
        html = u'<a href="{link}"><img src="{domain}{img_url}" title="{caption}" alt="{caption}"/></a>'\
            .format(link=link, domain=domain, img_url=banner.image.url,
                 caption=banner.caption)
        if autoescape:
            html = escape(html)
        return html

    @instance_cache
    def get_site(self, request=None):
        return get_current_site(request)

    def create_payment_request(self):
        payreq_model = self.pay_requests.model
        payreq_model(affiliate=self, amount=self.balance).save()

    def payed_to_affiliate(self, value):
        if self.balance < value:
            raise NotEnoughMoneyError()
        self.balance -= value
        self.total_payed += value

    def reward_affiliate(self, purchase_total_price, commit=True):
        """
        Rewards affiliate by incrementing its balance.
        :arg purchase_total_price: equals to total price of bought products
        :arg commit: if True, then commit changes to database, by calling save
        """
        self.total_payments_count += 1
        self.total_payments_amount += purchase_total_price
        if self.reward_percentage:
            reward = purchase_total_price * (self.reward_amount / D('100.0'))
            reward = self.quantize_amount(reward)
        else:
            reward = self.reward_amount
        self.balance += reward
        if commit:
            self.save()
        aff_stats = self.get_affiliate_stats()
        aff_stats.incr_payments(purchase_total_price, reward)
        affiliate_post_reward.send(sender=None, affiliate=self, reward=reward)

    def get_printable_reward(self):
        reward = D(self.reward_amount)
        if self.reward_percentage:
            return u"{0:.0f} %".format(reward)
        else:
            return u"{0:.0f} {1}".format(reward, self.get_currency())

    @staticmethod
    def quantize_amount(amount):
        return amount.quantize(TWOPLACES)

    def get_affiliate_stats(self):
        try:
            return self.stats.order_by("-date")[0]
        except IndexError:
            return self.stats.model(affiliate=self)

    @classmethod
    def create_affiliate(cls, *args, **kwargs):
        # Override this method to define your custom creation logic
        raise NotImplementedError()

    def get_currency(cls):
        # Override this method to define you currency label
        return settings.DEFAULT_CURRENCY


class AbstractAffiliateBanner(models.Model):
    image = models.ImageField(_("Banner image"),
        upload_to=BANNER_FOLDER)
    caption = models.CharField(_("Caption"), max_length=100)
    enabled = models.BooleanField(_("Enabled"), default=True)

    objects = AffiliateBannerManager()

    def __unicode__(self):
        return self.image.url

    class Meta:
        abstract = True
        verbose_name = _("Affiliate banner")
        verbose_name_plural = _("Affiliate banners")


class AbstractWithdrawRequest(models.Model):
    PAY_STATUS = Choices(
        ('pending', _("Pending")),
        ('done', _("Done")),
        ('error', _("Error")),
    )

    affiliate = models.ForeignKey(settings.AFFILIATE_MODEL,
        verbose_name=_("Affiliate"), related_name='pay_requests')
    status = models.CharField(_("Status"), max_length=10,
        choices=PAY_STATUS, default=PAY_STATUS.pending)
    amount = models.DecimalField(_("Payed to affiliate"),
        max_digits=6, decimal_places=2, default=D("0.0"))
    created_at = models.DateTimeField(_("Created at"), auto_now_add=True)
    payed_at = models.DateTimeField(_("Payed at"), null=True, blank=True)

    objects = WithdrawRequestManager()

    class Meta:
        abstract = True
        ordering = "-status", "-payed_at"
        verbose_name = _("Withdraw request")
        verbose_name_plural = _("Withdraw requests")

    def __unicode__(self):
        return u"{0} {1}".format(self.affiliate, self.status)

    def mark_done(self, commit=True):
        self.status = self.PAY_STATUS.done
        if commit:
            self.save()

    def payment_made(self):
        self.affiliate.payed_to_affiliate(self.amount)
        self.affiliate.save()
        self.payed_at = now()
        self.mark_done()
        affiliate_post_withdraw.send(sender=None, payment_request=self)

    def is_done(self):
        return self.status == self.PAY_STATUS.done
